import {
  CoreExecutorResult,
  CoreOperation,
  CoreResource,
  CoreResourceReader,
  CreateNewIdentifier,
} from "../../core";
import { PimCreateAssociation, PimCreateAssociationResult } from "../operation";
import { PimAssociation, PimAssociationEnd, PimClass } from "../model";
import { loadPimSchema, PimExecutorResultFactory } from "./pim-executor-utils";

export async function executesPimCreateAssociation(
  reader: CoreResourceReader,
  createNewIdentifier: CreateNewIdentifier,
  operation: CoreOperation
): Promise<CoreExecutorResult> {
  if (!PimCreateAssociation.is(operation)) {
    return PimExecutorResultFactory.invalidOperation();
  }

  const left = new PimAssociationEnd(createNewIdentifier("association-end"));
  left.pimPart = operation.pimAssociationEnds[0];
  const leftResult = await verityAssociationEnd(
    reader,
    operation,
    left.pimPart
  );
  if (leftResult?.failed) {
    return leftResult;
  }

  const right = new PimAssociationEnd(createNewIdentifier("association-end"));
  right.pimPart = operation.pimAssociationEnds[1];
  const rightResult = await verityAssociationEnd(
    reader,
    operation,
    right.pimPart
  );
  if (rightResult?.failed) {
    return leftResult;
  }

  const iri = operation.pimNewIri ?? createNewIdentifier("association");
  const result = new PimAssociation(iri);
  result.pimInterpretation = operation.pimInterpretation;
  result.pimTechnicalLabel = operation.pimTechnicalLabel;
  result.pimHumanLabel = operation.pimHumanLabel;
  result.pimHumanDescription = operation.pimHumanDescription;
  result.pimEnd = [left.iri, right.iri];
  result.pimIsOriented = operation.pimIsOriented;

  if (operation.pimAssociationEnds.length !== 2) {
    return CoreExecutorResult.createError(
      "Invalid number of ends for an association. Expected " +
        "2 given" +
        operation.pimAssociationEnds.length +
        "."
    );
  }

  const schema = await loadPimSchema(reader);
  if (schema === null) {
    return PimExecutorResultFactory.missingSchema();
  }

  return CoreExecutorResult.createSuccess(
    [result, left, right],
    [
      {
        ...schema,
        pimParts: [...schema.pimParts, result.iri, left.iri, right.iri],
      } as CoreResource,
    ],
    [],
    new PimCreateAssociationResult(result.iri, [left.iri, right.iri])
  );
}

async function verityAssociationEnd(
  reader: CoreResourceReader,
  operation: PimCreateAssociation,
  iri: string
): Promise<CoreExecutorResult | null> {
  const resource = await reader.readResource(iri);
  if (resource === null) {
    return PimExecutorResultFactory.missing(iri);
  }
  if (!PimClass.is(resource)) {
    return PimExecutorResultFactory.invalidType(resource, "pim:class");
  }
  return null;
}
